#!/usr/bin/env python3

"""Load configuration from file(s)."""

import os
import sys

from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)


def set_common_values(group_options, common_options):
    """Set common options across all option groups."""
    for key, value in common_options.items():
        if key not in group_options:
            group_options[key] = value
        elif isinstance(group_options[key], dict):
            for sub_key, sub_value in common_options[key].items():
                if sub_key not in group_options[key]:
                    group_options[key][sub_key] = sub_value


def update_subsets(flattened_group_options, group_options):
    """Update nested config options."""
    for key, value in group_options.items():
        if isinstance(group_options[key], dict):
            for sub_key, sub_value in group_options[key].items():
                if isinstance(group_options[key][sub_key], dict):
                    for sub_sub_key, sub_sub_value in group_options[key][
                        sub_key
                    ].items():
                        if isinstance(sub_sub_value, str) and sub_sub_value.startswith(
                            "~"
                        ):
                            sub_sub_value = os.path.expanduser(sub_sub_value)
                        elif (
                            not isinstance(sub_sub_value, list)
                            and sub_sub_key == "root"
                        ):
                            sub_sub_value = [sub_sub_value]
                        flattened_group_options[
                            "%s-%s-%s" % (key, sub_key, sub_sub_key)
                        ] = sub_sub_value
                else:
                    if isinstance(sub_value, str) and sub_value.startswith("~"):
                        sub_value = os.path.expanduser(sub_value)
                    flattened_group_options["%s-%s" % (key, sub_key)] = sub_value
        else:
            if isinstance(value, str) and value.startswith("~"):
                value = os.path.expanduser(value)
            flattened_group_options[key] = value


def load_config(options, file):
    """Load configuration from file."""
    new_options = tofile.load_yaml(file)
    if "common" in new_options:
        for group_name, group_options in new_options.items():
            if group_name != "common":
                set_common_values(group_options, new_options["common"])
    for group_name, group_options in new_options.items():
        if group_name not in options:
            options[group_name] = {}
        nested_group_options = options[group_name]
        update_subsets(nested_group_options, group_options)
    return options


def enforce_lists(options):
    """Ensure specified options are lists."""
    keys = {"es-host", "blank"}
    for key in keys:
        if key in options:
            if not isinstance(options[key], list):
                options[key] = [options[key]]


def set_taxonomy(options, group):
    """Convert taxonomy options to match allowed command line flags."""
    tax_opts = {
        "source": None,
        "format": None,
        "root": None,
        "file": None,
        "url": None,
        "path": None,
        "jsonl": None,
    }
    redundant_options = []
    source = options[group].get("taxonomy-source", None)
    for key, value in options[group].items():
        if key.startswith("taxonomy"):
            parts = key.split("-")
            if len(parts) == 3:
                redundant_options.append(key)
                if source is not None and parts[1] != source:
                    continue
                tax_opts.update({"format": parts[1], "source": parts[1]})
                if parts[2] in tax_opts:
                    tax_opts.update({parts[2]: value})
                else:
                    LOGGER.error("%s is not a valid option", key)
                    sys.exit(1)
            elif len(parts) == 2:
                if parts[1] in tax_opts:
                    tax_opts.update({parts[1]: value})
                else:
                    LOGGER.error("%s is not a valid option", key)
                    sys.exit(1)
            else:
                LOGGER.error("%s is not a valid option", key)
                sys.exit(1)
    for opts in options.values():
        redundant_opts = redundant_options[:]
        for key in opts.keys():
            if (
                key.startswith("taxonomy")
                and len(key.split("-")) == 3
                and key not in redundant_opts
            ):
                redundant_opts.append(key)
        for key, value in tax_opts.items():
            if value is not None:
                opts.update({"taxonomy-%s" % key: value})
        for key in redundant_opts:
            opts.pop(key, None)


def config(group, **kwargs):
    """Load configuration."""
    LOGGER.info("Loading configuration options")
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    dist_file = os.path.join(script_dir, "config", "dist.config.yaml")
    user_file = os.path.join(script_dir, "config", "config.yaml")
    options = {}
    if os.path.exists(dist_file):
        options = load_config(options, dist_file)
    else:
        print(dist_file)
    if os.path.exists(user_file):
        options = load_config(options, user_file)
    if "--config-file" in kwargs:
        for file in kwargs["--config-file"]:
            if os.path.exists(file):
                try:
                    options = load_config(options, file)
                except AttributeError:
                    LOGGER.error("Unable to parse config file %s", file)
                    sys.exit(1)
                except TypeError:
                    LOGGER.error("Unable to parse config file %s", file)
                    sys.exit(1)
            else:
                LOGGER.error("Config file %s could not be found", file)
                sys.exit(1)
    if group not in options:
        options[group] = {}
    for k, v in kwargs.items():
        if v:
            k = k.lstrip("--")
            if isinstance(v, str):
                if v.startswith("~"):
                    v = os.path.expanduser(v)
            if k in options[group]:
                options[group][k] = v
            elif not k == group:
                options[group][k] = v
    enforce_lists(options)
    set_taxonomy(options, group)
    return options



#!/usr/bin/env python3

"""Geographic methods."""

import re


def degrees_to_decimal(coord):
    """Convert latitude or longitude string to decimal."""
    if coord is None:
        return None
    try:
        coord = str(coord)
        multiplier = -1 if coord[-1] in ["S", "W"] else 1
        coord = re.sub(r"[NESW]", "", coord)
        parts = re.split(r"[°ʹ″]", coord)
        if len(parts) >= 3:
            return multiplier * sum(float(x) / 60**n for n, x in enumerate(parts[:3]))
        return multiplier * float(coord)
    except Exception:
        return None



#!/usr/bin/env python3
"""Hub functions."""


import contextlib
import csv
import json
import os
import re
import shutil
import sys
from collections import defaultdict
from copy import deepcopy
from operator import add
from operator import mod
from operator import mul
from operator import pow
from operator import sub
from operator import truediv
from pathlib import Path

from h3 import latlng_to_cell
from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)
MIN_INTEGER = -(2**31)
MAX_INTEGER = 2**31 - 1
DATE = re.compile(r"^[12]\d{3}-[01]\d-[0123]\d$")


def setup(opts):
    """Set up directory for this GenomeHubs instance and handle reset."""
    Path(opts["hub-path"]).mkdir(parents=True, exist_ok=True)
    return True


def dep(arg):
    """Dependency resolver."""
    # from https://code.activestate.com/recipes/576570-dependency-resolver/
    # © 2008 Louis Riviere (MIT)
    d = dict((k, set(arg[k])) for k in arg)
    r = []
    c = 0
    while d and c < 10:
        # values not in keys (items without dep)
        t = set(i for v in d.values() for i in v) - set(d.keys())
        # and keys without value (items without dep)
        t.update(k for k, v in d.items() if not v)
        # can be done right away
        r.append(t)
        # and cleaned up
        d = dict(((k, v - t) for k, v in d.items() if v))
        c += 1
    if d:
        LOGGER.error("Unable to resolve dependency tree")
        sys.exit(1)
    return r


def list_files(dir_path, pattern):
    """List files and sort by dependencies."""
    LOGGER.info("Finding files in %s matching %s", dir_path, pattern)
    deps = {}
    file_list = []
    yaml_files = sorted(Path(dir_path).glob(pattern))
    for yaml_file in yaml_files:
        yaml_dir = yaml_file.parent
        yaml_file = str(yaml_file)
        data = tofile.load_yaml(yaml_file)
        if "file" in data and "needs" in data["file"]:
            needs = data["file"]["needs"]
            if not isinstance(needs, list):
                needs = [needs]
            deps[yaml_file] = [f"{yaml_dir}/{needed}" for needed in needs]
        else:
            deps[yaml_file] = []
    file_list = [item for sublist in dep(deps) for item in sublist]
    return file_list


def copy_types(name, directory):
    """Copy a types file from the templates directory."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    with contextlib.suppress(Exception):
        types_file = os.path.join(script_dir, "templates", name)
        if not os.path.exists(f"directory/{name}"):
            shutil.copy(types_file, directory)


def load_types(name, *, part="types"):
    """Read a types file from the templates directory."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    types = None
    with contextlib.suppress(Exception):
        types_file = os.path.join(script_dir, "templates", f"{name}.{part}.yaml")
        types = tofile.load_yaml(types_file)
        # TODO: check if this code is redundant
        with contextlib.suppress(AttributeError):
            for key, value in types.items():
                with contextlib.suppress(KeyError):
                    enum = {
                        str(option).lower() for option in value["constraint"]["enum"]
                    }
                    value["constraint"]["enum"] = enum
    if types and "file" in types:
        for key, value in types["file"].items():
            if "attributes" in types:
                for attr in types["attributes"].values():
                    if (
                        isinstance(attr, (dict, list))
                        and key
                        not in ("exclusions", "format", "header", "name", "needs")
                        and key not in attr
                    ):
                        attr[key] = value
            if "taxon_names" in types:
                for attr in types["taxon_names"].values():
                    if (
                        key not in ("exclusions", "format", "header", "name")
                        and key not in attr
                    ):
                        attr[key] = value
    return types


def index_templator(parts, opts):
    """Index template helper function."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    mapping_file = os.path.join(script_dir, "templates", f"{parts[0]}.json")
    types = load_types(parts[0])
    names = load_types(parts[0], part="names")
    return {
        "name": parts[0],
        "index_name": opts["hub-separator"].join(parts),
        "mapping": tofile.load_yaml(mapping_file),
        "types": types,
        "names": names,
    }


def add_names_to_types(names, types):
    """Add names field meta to type field meta."""
    sources = 0
    if types is not None:
        types = deepcopy(types)
        sources += 1
    elif names is not None:
        types = deepcopy(names)
        sources += 1
    if sources == 2:
        for group, entries in names.items():
            if group not in types:
                types[group] = deepcopy(entries)
            else:
                for field, attrs in entries.items():
                    if isinstance(attrs, dict):
                        if field not in types[group]:
                            types[group][field] = deepcopy(attrs)
                        elif types[group][field]["header"] != attrs["header"]:
                            types[group][f"names_{field}"] = deepcopy(attrs)
    return types


def order_parsed_fields(parsed, types, names=None):
    """Order parsed fields using a template file."""
    columns = {}
    fields = {}
    ctr = 0
    types = add_names_to_types(names, types)
    for group, entries in types.items():
        for field, attrs in entries.items():
            header = False
            with contextlib.suppress(AttributeError):
                for key, value in attrs.items():
                    if key == "header":
                        header = value
                    elif key == "index":
                        if value not in columns:
                            columns[value] = field
                            fields[field] = value
                if header:
                    if not isinstance(header, list):
                        header = [header]
                    for head in header:
                        if head not in fields:
                            columns[ctr] = head
                            fields[head] = ctr
                            ctr += 1
    order = [x[0] for x in sorted(fields.items(), key=lambda x: x[1])]
    data = [order]
    for entry in parsed:
        row = [entry.get(field, "None") for field in order]
        data.append(row)
    return data


def post_search_scripts(es):
    """POST ElasticSearch search scripts."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    search_script_dir = os.path.join(script_dir, "templates", "scripts")
    directory = os.fsencode(search_script_dir)
    for script_file in os.listdir(directory):
        filename = os.fsdecode(script_file)
        if filename.endswith(".json"):
            body = tofile.load_yaml(os.path.join(search_script_dir, filename))
            script_id = filename.replace(".json", "")
            with tolog.DisableLogger():
                es.put_script(id=script_id, body=body)


def byte_type_constraint(value):
    """Test byte_type constraint."""
    return value >= -128 and value <= 128


def integer_type_constraint(value):
    """Test integer_type constraint."""
    return value >= MIN_INTEGER and value <= MAX_INTEGER


def short_type_constraint(value):
    """Test short_type constraint."""
    return value >= -32768 and value <= 32767


def min_value_constraint(value, limit):
    """Test minimum value constraint."""
    return value >= limit


def max_value_constraint(value, limit):
    """Test maximum value constraint."""
    return value <= limit


def enum_constraint(value, enum):
    """Test value in predefined set constraint."""
    return str(value).lower() in enum


def date_constraint(value, enum):
    """Test date constraint."""
    return DATE.match(value)


def test_constraint(value, constraint):
    """Test value against constraint."""
    constraints = {
        "byte": byte_type_constraint,
        "integer": integer_type_constraint,
        "short": short_type_constraint,
        "min": min_value_constraint,
        "max": max_value_constraint,
        "enum": enum_constraint,
        "date": date_constraint,
    }
    return not any(
        key in constraints and not constraints[key](value, constraint[key])
        for key in constraint
    )


def convert_lat_lon(location, *, hex=False, tile=False, resolution=3):
    """Convert lat and lon to array notation.

    Optionally convert to hex or tile format.

    Args:
        location (str): Latitude and longitude in string format.
        hex (bool): Convert to hex format.
        tile (bool): Convert to tile format.
        resolution (int): Resolution for tile format.
    Returns:
        str: Converted latitude and longitude in string format.
    """
    if location is None or not location:
        return ""
    sign = {"N": "", "E": "", "S": "-", "W": "-"}
    if not isinstance(location, list):
        if "," in location:
            location = re.split(r"\s*,\s*", location)
        elif location.endswith(("E", "W")):
            parts = re.split(r"\s*([NESW])\s*", location)
            location = [f"{sign[parts[1]]}{parts[0]}", f"{sign[parts[3]]}{parts[2]}"]
        else:
            return None
    for index, coord in enumerate(location):
        if coord.endswith(("N", "S", "E", "W")):
            parts = re.split(r"\s*([NESW])\s*", coord)
            location[index] = f"{sign[parts[1]]}{parts[0]}"
        if not re.match(r"^-*\d{1,3}\.*\d*$", location[index]):
            return None
    if abs(float(location[0])) > 90 or abs(float(location[1])) > 180:
        return None
    if hex:
        # use h3 library to convert lat/lon to hex
        return latlng_to_cell(float(location[0]), float(location[1]), resolution)
    if tile:
        # convert lat/lon to geo tile
        lat = float(location[0])
        lon = float(location[1])
        lat_rad = lat * (3.141592653589793 / 180)
        n = 2.0**resolution
        x_tile = int((lon + 180.0) / 360.0 * n)
        y_tile = int(
            (1.0 - (1.0 / 3.141592653589793) * (3.141592653589793 + lat_rad)) / 2.0 * n
        )
        x_tile = min(max(x_tile, 0), n - 1)
        y_tile = min(max(y_tile, 0), n - 1)
        location = [str(x_tile), str(y_tile)]
    return ",".join(location)


def convert_to_type(key, raw_value, to_type, *, translate=None):
    """Convert values to type."""
    # print(raw_values)
    # if not isinstance(raw_values, list):
    #     raw_values = [raw_values]
    # values = []
    # for raw_value in raw_values:
    if to_type in {"byte", "integer", "long", "short"}:
        try:
            value = int(raw_value)
        except ValueError:
            raw_value = str(raw_value).lower()
            if translate and raw_value in translate:
                new_value = translate[raw_value]
                if isinstance(new_value, list):
                    value = [convert_to_type(key, v, to_type) for v in new_value]
                else:
                    value = convert_to_type(key, translate[raw_value], to_type)
            else:
                value = None
    elif to_type in {
        "double",
        "float",
        "half_float",
        "1dp",
        "2dp",
        "3dp",
        "4dp",
    }:
        try:
            value = float(raw_value)
        except ValueError:
            raw_value = str(raw_value).lower()
            if translate and raw_value in translate:
                value = convert_to_type(key, translate[raw_value], to_type)
            else:
                value = None
        except TypeError:
            LOGGER.warn("Unable to import %s", key)
            value = None
    elif to_type == "geo_point":
        value = convert_lat_lon(raw_value)
    elif to_type == "geo_hex":
        value = convert_lat_lon(raw_value, hex=True)
    # elif to_type == "geo_tile":
    #     value = convert_lat_lon(raw_value, tile=True)
    elif to_type == "flattened":
        value = json.loads(raw_value)
    else:
        value = str(raw_value)
    if value is None:
        LOGGER.warning(
            "%s value %s is not a valid %s",
            key,
            str(raw_value),
            to_type,
        )
    #     if isinstance(value, list):
    #         values += value
    #     else:
    #         values.append(value)
    #     print(values)
    # if len(values) == 1:
    #     return values[0]
    return value


def calculate(string):
    """Recursively apply operations to a string."""
    operators = {
        "+": add,
        "-": sub,
        "%": mod,
        "*": mul,
        "/": truediv,
        "**": pow,
    }
    string = string.replace(" ", "")
    with contextlib.suppress(ValueError):
        return float(string)
    parts = re.split(r"(\(.+\))", string)
    if len(parts) > 1:
        for index, part in enumerate(parts):
            if part.startswith("("):
                part = part[1:-1]
                parts[index] = str(calculate(part))
        string = "".join(parts)
    for function in operators.keys():
        left, operator, right = string.partition(function)
        if operator in operators:
            try:
                return operators[operator](calculate(left), calculate(right))
            except TypeError:
                return None


def attribute_value_query(es, index, opts=None):
    """Run attribute value query."""
    query = {
        "bool": {
            "filter": [
                {"match": {opts["id_field"]: opts["primary_id"]}},
                {
                    "nested": {
                        "path": "attributes",
                        "query": {
                            "bool": {
                                "filter": [
                                    {"match": {"attributes.key": opts["attribute"]}}
                                ]
                            }
                        },
                        "inner_hits": {
                            "_source": False,
                            "name": "%s_values" % opts["attribute"],
                            "docvalue_fields": [
                                "attributes.key",
                                "attributes.%s_value" % opts["value_type"],
                            ],
                        },
                    }
                },
            ]
        }
    }
    return es.search(index=index, body={"query": query, "_source": False})


def lookup_attribute_value(identifier, attribute, shared_values):
    """Lookup an indexed attribute value."""
    value_type = shared_values["_types"]["attributes"][attribute]["type"]
    opts = {
        "id_field": f'{shared_values["_index_type"]}_id',
        "primary_id": identifier,
        "attribute": attribute,
        "value_type": value_type,
    }
    res = attribute_value_query(
        shared_values["_es"],
        shared_values["_index"],
        opts,
    )
    hits = res["hits"]["hits"]
    try:
        if len(hits) == 1:
            inner_hits = hits[0]["inner_hits"][f"{attribute}_values"]["hits"]["hits"]
            if len(inner_hits) == 1:
                field_values = inner_hits[0]["fields"][f"attributes.{value_type}_value"]
                if len(field_values) == 1:
                    shared_values[identifier][attribute] = field_values[0]
                    return field_values[0]
    except KeyError:
        print(inner_hits)
        sys.exit(1)
    return None


def apply_template(value, operation, row_values, shared_values):
    """Apply template to a function description."""
    parts = re.split(r"(\{.*?\})", operation)
    for index, part in enumerate(parts):
        if part.startswith("{"):
            part = part[1:-1]
            if not part:
                parts[index] = value
            elif part in row_values:
                parts[index] = str(row_values[part])
            elif part.startswith("."):
                part = part[1:]
                if value in shared_values and part in shared_values[value]:
                    parts[index] = str(shared_values[value][part])
                else:
                    parts[index] = str(
                        lookup_attribute_value(value, part, shared_values)
                    )
                if parts[index] is None:
                    LOGGER.error("%s has no value for attribute %s", value, part)
                    sys.exit()
            else:
                LOGGER.error("function template '%s' is not supported", operation)
                sys.exit(1)
    return "".join(parts)


def calculator(value, operation, row_values, shared_values, template_type):
    """Template-based calculator."""
    # print(dict(shared_values))
    operation = apply_template(value, operation, row_values, shared_values)
    return operation if template_type == "template" else calculate(operation)


def validate_values(values, key, types, row_values, shared_values, blanks):
    """Validate values."""
    validated = []
    if isinstance(types[key], str) or "type" not in types[key]:
        types[key]["type"] = "keyword"
    key_type = types[key]["type"]
    for value in values:
        if value in blanks:
            continue
        if "function" in types[key] or "template" in types[key]:
            if re.match(r"^\d+\.\d+e[\+-]\d+$", value):
                value = str(float(value))
            try:
                template_type = "function" if "function" in types[key] else "template"
                value = calculator(
                    value,
                    types[key][template_type],
                    row_values,
                    shared_values,
                    template_type,
                )
            except ValueError:
                continue
        value = convert_to_type(
            key, value, key_type, translate=types[key].get("translate", None)
        )
        if isinstance(types[key], str):
            validated.append(value)
            continue
        if value is None:
            continue
        if isinstance(value, str):
            if not value:
                continue
            with contextlib.suppress(KeyError):
                value = types[key]["translate"][value.lower()]
        if not isinstance(value, list):
            value = [value]
        for v in value:
            if not isinstance(v, dict) and v in blanks:
                continue
            try:
                valid = test_constraint(v, key_type)
                if valid:
                    valid = test_constraint(v, types[key]["constraint"])
            except KeyError:
                valid = True
            if valid:
                validated.append(v)
            elif v:
                LOGGER.warning("%s is not a valid %s value", str(v), key)
    return validated


def apply_value_template(prop, value, attribute, *, taxon_types, has_taxon_data):
    """Set value using template."""
    template = re.compile(r"^(.*?\{\{)(.+)(\}\}.*)$")
    new_prop = prop.replace("taxon_", "")
    if match := template.match(str(value)):
        groups = match.groups()
        if groups[1] and groups[1] in attribute:
            has_taxon_data = True
            new_value = attribute[groups[1]]
            if groups[0] != "{{":
                new_value = "%s%s" % (
                    groups[0].replace("{{", ""),
                    str(new_value),
                )
            if groups[2] != "}}":
                new_value = "%s%s" % (
                    str(new_value),
                    groups[2].replace("}}", ""),
                )
            update_taxon_types(taxon_types, new_prop, new_value, prop)
    else:
        has_taxon_data = True
        update_taxon_types(taxon_types, new_prop, value, prop)
    return has_taxon_data


def update_taxon_types(taxon_types, new_prop, value, prop):
    """Set taxon attribute property."""
    taxon_types.update({new_prop: value, "group": "taxon"})
    if prop in taxon_types:
        del taxon_types[prop]


def process_metadata(attribute, blanks):
    """Move attribute metadata to a single dict."""
    paths = {
        "comment": "value.comment",
        "is_primary_value": "value.primary",
        "organelle": "value.organelle",
        "source": "source.name",
        "source_pubmed_id": "source.pubmed",
        "source_url_stub": "source.stub",
        "url": "source.url",
    }
    metadata = attribute["metadata"] if "metadata" in attribute else {}
    if not isinstance(metadata, dict):
        metadata = json.loads(metadata)
    for key, value in attribute.items():
        if key in paths:
            path = paths[key]
        elif key.startswith("source_"):
            path = key.replace("_", ".")
        else:
            continue
        parent = metadata
        parts = path.lower().split(".")
        for index, part in enumerate(parts):
            if index < len(parts) - 1:
                if part not in parent:
                    parent[part] = {}
                parent = parent[part]
            elif part not in parent or parent[part] in blanks:
                parent[part] = str(value).lower()
    attribute["metadata"] = metadata


def add_attributes(
    entry,
    types,
    *,
    attributes=None,
    source=None,
    attr_type="attributes",
    meta=None,
    shared_values=None,
    row_values=None,
    blanks=None,
):
    """Add attributes to a document."""
    (
        attributes,
        meta,
        row_values,
        taxon_attributes,
        taxon_types,
        attribute_values,
    ) = set_add_attributes_defaults(attributes, meta, row_values)
    indices = {}
    for k, values in entry.items():
        key, *rest = k.split(".") if "." in k else [k]
        if rest:
            try:
                attribute = attributes[indices[key]]
            except KeyError:
                continue
            if "metadata" not in attribute:
                attribute["metadata"] = {}
            md = attribute["metadata"]
            if (isinstance(values, str) and values in blanks) or any(
                value in blanks for value in values
            ):
                continue
            for index, part in enumerate(rest):
                if part not in md:
                    md[part] = {}
                if index < len(rest) - 1:
                    md = md[part]
                else:
                    md[part] = values
        elif key in types:
            if not isinstance(values, list):
                values = [values]
            if attr_type == "taxon_names":
                validated = values
            else:
                validated = validate_values(
                    values, key, types, row_values, shared_values, blanks
                )
            # TODO: handle invalid values
            if validated:
                if len(validated) == 1:
                    validated = validated[0]
                row_values[key] = validated
                if attr_type == "attributes":
                    attribute = {"key": key, f'{types[key]["type"]}_value': validated}
                    attribute_values[key] = attribute
                elif attr_type == "taxon_names":
                    attribute = {"name": validated, "class": key, **types[key]}
                else:
                    attribute = {"identifier": validated, "class": key, **types[key]}
                if "source" in types[key]:
                    attribute.update(meta)
                    attribute.update(
                        {
                            k: v
                            for k, v in types[key].items()
                            if k.startswith("source") and k not in attribute
                        }
                    )
                else:
                    attribute.update(meta)
                    if "source" not in attribute and source is not None:
                        attribute.update({"source": source})
                if "source_url_template" in attribute:
                    attribute["source_url"] = apply_template(
                        validated,
                        attribute["source_url_template"],
                        row_values,
                        shared_values,
                    )
                # if "metadata" in types[key]:
                #     if "metadata" not in attribute:
                #         attribute.update({"metadata": {}})
                #     attribute.update(
                #         {
                #             "metadata": {
                #                 k: v
                #                 for k, v in types[key]["metadata"].items()
                #                 if k not in attribute["metadata"]
                #             }
                #         }
                #     )
                indices[key] = len(attributes)
                attributes.append(attribute)
    data_freeze = False
    if attribute_values:
        for attribute in attributes:
            process_metadata(attribute, blanks)
            if (
                attribute.get("key", "") == "data_freeze"
                and attribute.get("keyword_value", "latest") != "latest"
            ):
                data_freeze = True
            has_taxon_data = False
            taxon_attribute_types = {**types[attribute["key"]]}
            taxon_attribute = {**attribute, "source_index": "assembly"}
            # taxon_props = []
            for prop, value in types[attribute["key"]].items():
                if prop.startswith("taxon_"):
                    has_taxon_data = apply_value_template(
                        prop,
                        value,
                        attribute,
                        taxon_types=taxon_attribute_types,
                        has_taxon_data=has_taxon_data,
                    )
                    # taxon_props.append(prop)
            # for prop in taxon_props:
            #     del types[attribute["key"]][prop]
            if has_taxon_data:
                if "name" not in taxon_attribute_types:
                    taxon_attribute_types["name"] = attribute["key"]
                taxon_attribute["key"] = taxon_attribute_types["name"]
                # taxon_attribute.update({"name": taxon_attribute_types["key"]})
                taxon_attributes.append(taxon_attribute)
                taxon_types[taxon_attribute_types["name"]] = taxon_attribute_types
    if data_freeze:
        taxon_attributes = []
    return attributes, taxon_attributes, taxon_types


def set_add_attributes_defaults(attributes, meta, row_values):
    """Set default values for add_attributes."""
    if attributes is None:
        attributes = []
    taxon_attributes = []
    taxon_types = {}
    if meta is None:
        meta = {}
    attribute_values = {}
    if row_values is None:
        row_values = {}
    return attributes, meta, row_values, taxon_attributes, taxon_types, attribute_values


def chunks(arr, n):
    """Yield successive n-sized chunks from arr."""
    for i in range(0, len(arr), n):
        yield arr[i : i + n]


def add_attribute_values(existing, new, *, raw=True):
    """Add attribute values to records."""
    indices = {}
    for index, entry in enumerate(existing):
        indices[entry["key"]] = index
    index = len(existing)
    for group in new:
        if not isinstance(group, list):
            group = [group]
        for entry in group:
            if raw:
                arr = []
                if indices and entry["key"] in indices:
                    try:
                        arr = existing[indices[entry["key"]]]["values"]
                    except KeyError:
                        LOGGER.error(
                            "Unable to import values to an attribute (%s) that has already been filled",
                            entry["key"],
                        )
                        exit(1)
                else:
                    existing.append({"key": entry["key"], "values": arr})
                    indices[entry["key"]] = index
                    index += 1
                del entry["key"]
                arr.append(entry)
            else:
                mod_entry = {**entry, "count": 1}
                if "keyword_value" in entry:
                    if isinstance(mod_entry["keyword_value"], list):
                        mod_entry["length"] = len(mod_entry["keyword_value"])
                    else:
                        mod_entry["length"] = 1
                existing.append(mod_entry)


def strip_comments(data, types):
    """Strip comment lines from a file stream."""
    comment_chars = {"#"}
    if "file" in types and "comment" in types["file"]:
        comment_chars.update(set(types["file"]["comment"]))
    for row in data:
        if row[0] in comment_chars:
            continue
        yield row


def process_names_file(types, names_file, *, value_path=None):
    """Process a taxon names file."""
    data = tofile.open_file_handle(names_file)
    names = defaultdict(dict)
    if data is None:
        return names
    delimiters = {"csv": ",", "tsv": "\t"}
    rows = csv.reader(
        strip_comments(data, types),
        delimiter=delimiters[types["file"]["format"]],
        quotechar='"',
    )
    if value_path and isinstance(value_path, dict):
        for row in rows:
            for group, keys in value_path.items():
                if not isinstance(keys, list):
                    keys = [keys]
                for key in keys:
                    if key not in names[group]:
                        names[group][key] = {row[0]: row[1]}
                    else:
                        names[group][key][row[0]] = row[1]
    else:
        next(rows)
        for row in rows:
            name = row[3] if len(row) > 3 else row[1]
            names[row[2]][row[1]] = {"name": name, "taxon_id": row[0]}
    return names


def set_xrefs(taxon_names, types, row, *, meta=None):
    """Set xrefs for taxon_names."""
    if meta is None:
        meta = {}
    names = []
    for taxon in taxon_names:
        if "xref" in types[taxon["class"]] and types[taxon["class"]]["xref"]:
            if "source" in meta:
                taxon.update({"source": meta["source"]})
            if "source_url_stub" in meta:
                taxon.update({"source_url_stub": meta["source_url_stub"]})
        names.append(taxon)
    return names


def set_row_defaults(types, data):
    """Set default values for a row."""
    for key in types["defaults"].keys():
        if key in types:
            for entry in types[key].values():
                if not isinstance(entry, dict):
                    entry = {"default": entry}
                entry = {
                    **types["defaults"][key],
                    **entry,
                }
        elif key == "metadata":
            data["metadata"] = {**types["defaults"]["metadata"]}


def process_row_values(row, types, data):
    """Process row values."""
    for group in data.keys():
        if group in types:
            for key, meta in types[group].items():
                if not isinstance(meta, dict):
                    data[group][key] = meta
                    continue
                if "index" not in meta:
                    if "default" in meta:
                        data[group][key] = meta["default"]
                    continue
                try:
                    if isinstance(meta["index"], list):
                        char = meta.get("join", "")
                        values = [row[i] for i in meta["index"]]
                        if all(values):
                            value = char.join(values)
                        else:
                            continue
                    else:
                        value = row[meta["index"]]
                    if "separator" in meta and any(
                        sep in value for sep in meta["separator"]
                    ):
                        separator = "|".join(
                            [re.escape(sep) for sep in meta["separator"]]
                        )
                        data[group][key] = re.split(rf"\s*{separator}\s*", value)
                        if "limit" in meta:
                            data[group][key] = data[group][key][: meta["limit"]]
                    elif value is not None and value != "None":
                        data[group][key] = value
                except IndexError:
                    LOGGER.warning(f"Missing fields in row '{str(row)}'")
                    return None
                except Exception as err:
                    LOGGER.warning(f"Cannot parse row '{str(row)}'")
                    raise err
    return True


def process_taxon_names(data, types, row, names):
    """Process taxon names."""
    if data["taxon_names"]:
        data["taxon_names"] = set_xrefs(
            data["taxon_names"], types["taxon_names"], row, meta=data["metadata"]
        )
    if data["taxonomy"] and names:
        for key in names.keys():
            if key in data["taxonomy"] and data["taxonomy"][key] in names[key]:
                data["taxonomy"]["_taxon_id"] = names[key][data["taxonomy"][key]][
                    "taxon_id"
                ]
                data["taxonomy"][key] = names[key][data["taxonomy"][key]]["name"]


def process_features(data):
    """Move feature properties to top level."""
    if data["features"]:
        for entry in data["features"]:
            data[entry["class"]] = entry["identifier"]
        del data["features"]


def contains_excluded_value(data, exclusions):
    """Check whether row data matches an excluded value."""
    for key in exclusions.keys():
        for subkey in exclusions[key].keys():
            if str(data.get(key, {}).get(subkey, "")) in exclusions[key][subkey]:
                return True


def process_row(
    types, names, row, shared_values, blanks, *, index_type="assembly", exclusions=None
):
    """Process a row of data."""
    data = {
        "attributes": {},
        "features": {},
        "identifiers": {},
        "metadata": {},
        "taxon_names": {},
        "taxonomy": {},
        "taxon_attributes": {},
    }
    set_row_defaults(types, data)
    if process_row_values(row, types, data) is None:
        return None, None, None
    if exclusions is None:
        exclusions = defaultdict(dict)
    if contains_excluded_value(data, exclusions):
        return None, None, None
    taxon_data = {}
    taxon_types = {}
    if "is_primary_value" in data["metadata"]:
        try:
            data["metadata"]["is_primary_value"] = bool(
                int(data["metadata"]["is_primary_value"])
            )
        except ValueError:
            data["metadata"]["is_primary_value"] = False
    row_id = None
    if (
        "identifiers" in data
        and data["identifiers"]
        and f"{index_type}_id" in data["identifiers"]
        and data["identifiers"][f"{index_type}_id"]
    ):
        row_id = data["identifiers"][f"{index_type}_id"]
    row_values = {}
    for attr_type in ["identifiers", "taxon_names", "attributes", "features"]:
        if attr_type in data and data[attr_type]:
            (
                data[attr_type],
                taxon_data[attr_type],
                taxon_types[attr_type],
            ) = add_attributes(
                data[attr_type],
                types[attr_type],
                attr_type=attr_type,
                meta=data["metadata"],
                shared_values=shared_values,
                row_values=row_values,
                blanks=blanks,
            )
        else:
            data[attr_type] = []
    if row_id is not None and "attributes" in taxon_data and taxon_data["attributes"]:
        for attr in taxon_data["attributes"]:
            attr.update(
                {
                    "source_index": index_type,
                    "source_id": row_id,
                }
            )
    process_taxon_names(data, types, row, names)
    process_features(data)
    return data, taxon_data, taxon_types.get("attributes", {})


def find_index_for_header(header, headers, duplicate_headers):
    """Lookup column index for a single header."""
    index = headers.get(header, None)
    if index is not None:
        if header in duplicate_headers:
            LOGGER.error(
                "Duplicate header '%s' in columns %s",
                header,
                " and ".join([str(idx) for idx in duplicate_headers[header]]),
            )
            sys.exit(1)
        return index
    else:
        LOGGER.error("Unable to find %s in file headers", header)
        sys.exit(1)


def set_column_indices(types, header):
    """Use header to set indices for named columns."""
    headers = {}
    duplicate_headers = defaultdict(set)
    for index, title in enumerate(header):
        if title in headers:
            duplicate_headers[title].update([headers[title], index])
        headers[title] = index
    headers = {title: index for index, title in enumerate(header)}
    for entries in types.values():
        for value in entries.values():
            if isinstance(value, dict) and "header" in value:
                if isinstance(value["header"], list):
                    # convert list of headers to list of indices
                    value.update(
                        {
                            "index": [
                                find_index_for_header(head, headers, duplicate_headers)
                                for head in value["header"]
                            ]
                        }
                    )
                else:
                    value.update(
                        {
                            "index": find_index_for_header(
                                value["header"], headers, duplicate_headers
                            )
                        }
                    )


def write_imported_rows(rows, opts, *, types, header=None, label="imported"):
    """Write imported rows to processed file."""
    file_key = f'{opts["index"]}-exception'
    dir_key = f'{opts["index"]}-dir'
    if file_key in opts and opts[file_key]:
        outdir = opts[file_key]
    else:
        outdir = f"{opts[dir_key]}/{label}"
    os.makedirs(outdir, exist_ok=True)
    outfile = f'{outdir}/{types["file"]["name"]}'
    data = []
    header_len = 0
    if header is not None:
        data.append(header)
        header_len = 1
    if isinstance(rows, dict):
        for row_set in rows.values():
            data.extend(iter(row_set))
    else:
        data.extend(iter(rows))
    LOGGER.info(
        "Writing %d records to %s file '%s'", len(data) - header_len, label, outfile
    )
    tofile.write_file(outfile, data)


def write_spellchecked_taxa(spellings, opts, *, types):
    """Write spellchecked taxa to file."""
    dir_key = f'{opts["index"]}-dir'
    filepath = Path(types["file"]["name"])
    extensions = "".join(filepath.suffixes)
    file_basename = str(filepath).replace(extensions, "")
    dirs = {
        "spellcheck": "exceptions",
        "synonym": "imported",
    }
    for group in dirs.keys():
        for group in dirs.keys():
            taxa = []
            for name, obj in spellings[group].items():
                taxa.append([obj["taxon_id"], name, obj["rank"]] + obj["matches"])
            if taxa:
                outdir = f"{opts[dir_key]}/{dirs[group]}"
                os.makedirs(outdir, exist_ok=True)
                outfile = f"{outdir}/{file_basename}.spellcheck.tsv"
                LOGGER.info(
                    "Writing %d %s suggestions to spellcheck file '%s'",
                    len(taxa),
                    group,
                    outfile,
                )
                tofile.write_file(
                    outfile, [["taxon_id", "input", "rank", "suggested"]] + taxa
                )


def write_imported_taxa(taxa, opts, *, types):
    """Write imported taxa to file."""
    imported = []
    file_key = f'{opts["index"]}-exception'
    dir_key = f'{opts["index"]}-dir'
    filepath = Path(types["file"]["name"])
    extensions = "".join(filepath.suffixes)
    file_basename = str(filepath).replace(extensions, "")
    for name, arr in taxa.items():
        prefix = "#" if len(arr) > 1 else ""
        for obj in arr:
            if obj.get("additional_taxon", False):
                prefix = "#"
            imported.append([f'{prefix}{str(obj["taxon_id"])}', name, obj["rank"]])
    if imported:
        if file_key in opts and opts[file_key]:
            outdir = opts[file_key]
        else:
            outdir = f"{opts[dir_key]}/imported"
        os.makedirs(outdir, exist_ok=True)
        outfile = f"{outdir}/{file_basename}.taxon_ids.tsv"
        LOGGER.info(
            "Writing %d taxon_ids to imported file '%s'",
            len(imported),
            outfile,
        )
        tofile.write_file(outfile, [["taxon_id", "input", "rank"]] + imported)


def do_replace(item, arr):
    """Recursively replace all instances of query string."""
    if isinstance(item, dict):
        matched_keys = {}
        for key, value in item.items():
            item[key] = do_replace(value, arr)
            matched_keys[key] = do_replace(key, arr)
        for key, new_key in matched_keys.items():
            if new_key != key:
                item[new_key] = item[key]
                del item[key]
    elif isinstance(item, list):
        item = [do_replace(value, arr) for value in item]
    elif isinstance(item, str):
        for query, replace in arr:
            new_item = item.replace(query, str(replace))
            if new_item != item:
                item = new_item
    return item


def deep_replace(obj, arr):
    """Replace all instances of each query string in obj."""
    new_obj = deepcopy(obj)
    new_obj = do_replace(new_obj, arr)
    return new_obj



#!/usr/bin/env python3

"""Taxonomy methods."""

import contextlib
import json
from pathlib import Path

from tolkein import tofetch
from tolkein import tolog
from tolkein import totax

from .hub import index_templator

LOGGER = tolog.logger(__name__)


def index_template(name, opts):
    """Index template (includes name, mapping and types)."""
    index_name = "taxonomy"
    parts = [index_name, name, opts["hub-name"], opts["hub-version"]]
    return index_templator(parts, opts)


def files_exist(expected_files, path):
    """Test if expected files already exist."""
    return all((path / filename).exists() for filename in expected_files)


def confirm_index_opts(taxonomy_name, opts):
    """Confirm expected keys are present in opts for indexing."""
    file_key = "taxonomy-file"
    url_key = "taxonomy-url"
    for key in {"taxonomy-format", "taxonomy-path", file_key}:
        if key not in opts:
            LOGGER.warning("Unable to index %s, '%s' not specified", taxonomy_name, key)
            return False

    file_key = "taxonomy-file"
    taxonomy_path = Path(f'{opts["taxonomy-path"]}')
    if not files_exist(opts[file_key], taxonomy_path):
        taxonomy_path = Path(f'{opts["taxonomy-path"]}/{taxonomy_name}')
    taxonomy_path.mkdir(parents=True, exist_ok=True)
    if url_key not in opts and not files_exist(opts[file_key], taxonomy_path):
        LOGGER.warning(
            "Unable to index %s, '%s' not specified and files not found at '%s'",
            taxonomy_name,
            url_key,
            str(taxonomy_path),
        )
        return False
    return True


def index(taxonomy_name, opts):
    """Index a taxonomy."""
    if not confirm_index_opts(taxonomy_name, opts):
        return
    LOGGER.info("Indexing %s", taxonomy_name)
    template = index_template(taxonomy_name, opts)
    file_key = "taxonomy-file"
    taxonomy_path = Path(f'{opts["taxonomy-path"]}')
    if not files_exist(opts[file_key], taxonomy_path):
        taxonomy_path = taxonomy_path / taxonomy_name

    jsonl_files = [f for f in opts.get(file_key, []) if f.endswith(".jsonl")]
    for jsonl_file in jsonl_files:
        nodes_file = taxonomy_path / jsonl_file
        if nodes_file.exists():
            LOGGER.info(
                "Using existing %s taxdump at %s", taxonomy_name, str(taxonomy_path)
            )

            def stream_nodes():
                with open(nodes_file, "r") as f:
                    for line in f:
                        # parse json and get taxon_id
                        data = json.loads(line)
                        taxon_id = data.get("taxon_id")
                        if taxon_id:
                            yield (f"taxon-{taxon_id}", line)

            stream_nodes()
            return template, stream_nodes()

    if not files_exist(opts[file_key], taxonomy_path):
        LOGGER.info(
            "Fetching %s taxdump and extracting to %s",
            taxonomy_name,
            str(taxonomy_path),
        )
        log = opts.get("log-fetch", True)
        if log is True:
            tofetch.fetch_tar(url=opts["taxonomy-url"], path=str(taxonomy_path))
        else:
            if log is False:
                log = "/dev/null"
            with open(log, "a") as f:
                with contextlib.redirect_stderr(f):
                    tofetch.fetch_tar(url=opts["taxonomy-url"], path=str(taxonomy_path))
        for file in opts[file_key]:
            for p in taxonomy_path.rglob(file):
                p.rename(taxonomy_path / p.name)
    else:
        LOGGER.info(
            "Using existing %s taxdump at %s", taxonomy_name, str(taxonomy_path)
        )
    root_key = "taxonomy-root"
    root = opts.get(root_key, None)
    stream = totax.parse_taxonomy(opts["taxonomy-format"], str(taxonomy_path), root)
    return template, stream



#!/usr/bin/env python3
"""Window functions."""

import os

from tolkein import tofile
from tolkein import tolog

from .hub import deep_replace

LOGGER = tolog.logger(__name__)


def parse_config(config):
    """Parse values from config file."""
    revision = config.get("revision", 0)
    blobtoolkit_id = config["assembly"]["prefix"]
    if revision:
        blobtoolkit_id += f".{revision}"
    lineages = config.get("busco", {}).get("lineages", [])
    libraries = config.get("reads", {}).get("paired", []) + config.get("reads", {}).get(
        "single", []
    )
    replacements = [
        ("<<accession>>", config["assembly"]["accession"]),
        ("<<blobtoolkit_id>>", blobtoolkit_id),
        ("<<taxon_id>>", config["taxon"]["taxid"]),
        ("<<level>>", config["assembly"]["level"]),
        ("<<span>>", config["assembly"]["span"]),
        ("<<lineage>>", lineages[0]),
    ]
    return replacements, libraries


def fill_window_template(config_file, types, size=None):
    """Replace template values in window types file."""
    config = tofile.load_yaml(config_file)
    replacements, libraries = parse_config(config)
    if size is not None:
        replacements.append(("<<window>>", size))
    types = deep_replace(types, replacements)
    if libraries:
        prefix = libraries[0]["prefix"]
        types["attributes"]["coverage"] = {
            "header": f"{prefix}_cov",
            "function": "{} + 0.01",
        }
    return types


def window_parser(params, opts, *, types=None, names=None):
    """Parse window_stats files."""
    config_file = (
        opts.get("window", False) or opts.get("window-full", False)
    ) + "/config.yaml"
    if os.path.exists(config_file):
        new_types = fill_window_template(
            config_file, types, opts.get("window-size", [None])[0]
        )
    for key, value in new_types.items():
        types[key] = value
    return None



#!/usr/bin/env python3

"""
Index a file, directory or repository.

Usage:
    genomehubs index [--hub-name STRING] [--hub-path PATH] [--hub-version PATH]
                     [--config-file PATH...] [--config-save PATH]
                     [--es-batch INT] [--es-host URL...] [--assembly-dir PATH]
                     [--feature-dir PATH] [--sample-dir PATH]
                     [--taxon-dir PATH] [--taxon-repo URL] [--taxon-exception PATH]
                     [--taxon-lookup STRING] [--taxon-lookup-root STRING]
                     [--taxon-lookup-in-memory] [--taxon-id-as-xref STRING]
                     [--taxon-matching-ranks INT]
                     [--taxon-spellcheck] [--taxonomy-source STRING]
                     [--blank STRING...]
                     [--file PATH...] [file-dir PATH...]
                     [--remote-file URL...] [--remote-file-dir URL...]
                     [--taxon-id STRING] [--assembly-id STRING]
                     [--sample-id STRING] [--analysis-id STRING]
                     [--file-title STRING] [--file-description STRING] [--file-metadata PATH]
                     [--dry-run] [--log-interval INT] [--log-es BOOL]
                     [-h|--help] [-v|--version]

Options:
    --hub-name STRING          GenomeHubs instance name.
    --hub-path PATH            GenomeHubs instance root directory path.
    --hub-version STR          GenomeHubs instance version string.
    --config-file PATH         Path to YAML file containing configuration options.
    --config-save PATH         Path to write configuration options to YAML file.
    --es-batch INT             Batch size for ElasticSearch bulk indexing.
    --es-host URL              ElasticSearch hostname/URL and port.
    --assembly-dir PATH        Path to directory containing assembly-level data.
    --sample-dir PATH          Path to directory containing sample-level data.
    --feature-dir PATH         Path to directory containing feature-level data.
    --taxon-lookup-root STRING Root taxon Id for in-memory lookup.
    --taxon-lookup STRING      Taxon name class to lookup (scientific|any). [Default: scientific]
    --taxon-lookup-in-memory   Flag to use in-memory taxon name lookup.
    --taxon-id-as-xref STRING  Set source DB name to treat taxon_id in file as xref.
    --taxon-matching-ranks INT Number of ancestral ranks that must match to import a taxon based on
                               name match. [Default: 2]
    --taxon-spellcheck         Flag to use fuzzy matching to match taxon names.
    --taxon-dir PATH           Path to directory containing taxon-level data.
    --taxon-repo URL           Remote git repository containing taxon-level data.
                               Optionally include `~branch-name` suffix.
    --taxon-exception PATH     Path to directory to write taxon data that failed to import.
    --taxonomy-source STRING   Name of taxonomy to use (ncbi or ott).
    --file PATH                Path to file for generic file import.
    --file-dir PATH            Path to directory containing generic files to import.
    --remote-file URL          Location of remote file for generic file import.
    --remote-file-dir URL      Location of remote directory containing generic files to import.
    --taxon-id STRING          Taxon ID to index files against.
    --sample-id STRING         Sample ID to index files against.
    --assembly-id STRING       Assembly ID to index files against.
    --analysis-id STRING       Analysis ID to index files against.
    --file-title STRING        Default title for indexed files.
    --file-description STRING  Default description for all indexed files.
    --file-metadata PATH       CSV, TSV, YAML or JSON file metadata with one entry per file to be indexed.
    --blank STRING...          List of strings to treat as blank values in files. Default: ['', 'NA', 'N/A', 'None']
    --dry-run                  Flag to run without loading data into the elasticsearch index.
    --log-interval INT         Minimum time (seconds) between prgress bar updates
    --log-es BOOL              Show Info-level logs from elasticsearch
    -h, --help                 Show this
    -v, --version              Show version number

Examples:
    # 1. Index all files in a remote repository
    ./genomehubs index --taxon-repo https://github.com/genomehubs/goat-data
"""

import csv
import sys

# import time
from collections import defaultdict
from pathlib import Path
from time import sleep
from traceback import format_exc

from docopt import docopt
from tolkein import tolog
from tqdm import tqdm

from ..lib import es_functions
from ..lib import feature
from ..lib import hub
from ..lib import taxon
from . import sample
from .attributes import index_types
from .config import config
from .es_functions import index_stream
from .files import index_files
from .files import index_metadata
from .hub import list_files
from .hub import process_row
from .hub import set_column_indices
from .hub import strip_comments
from .hub import write_imported_rows
from .hub import write_imported_taxa
from .hub import write_spellchecked_taxa
from .sample import add_identifiers_and_attributes_to_entries
from .taxon import add_names_and_attributes_to_taxa
from .taxon import fix_missing_ids
from .taxon import load_taxon_table
from .taxon import translate_xrefs
from .test import test_json_dir
from .validate import validate_types_file
from .version import __version__

LOGGER = tolog.logger(__name__)


def not_blank(key, obj, blanks):
    """Test value is not blank."""
    return key in obj and obj[key] and obj[key] not in blanks


def summarise_imported_taxa(docs, imported_taxa):
    """Summarise taxon imformation from a stram of taxon docs."""
    for entry_id, entry in docs:
        imported_taxa[entry["scientific_name"]].append(
            {
                "taxon_id": entry["taxon_id"],
                "rank": entry["taxon_rank"],
                "additional_taxon": entry.get("additional_taxon", False),
            }
        )
        yield entry_id, entry


def group_rows(
    taxon_id,
    rows,
    with_ids,
    without_ids,
    taxon_asm_data,
    imported_rows,
    types,
    failed_rows,
    blanks,
):
    """Group processed rows by available taxon info for import."""
    ranks = ["subspecies", "species", "family", "order", "class"]
    for processed_data, taxon_data, row in rows:
        if taxon_id != "other":
            with_ids[taxon_id].append(processed_data)
            taxon_asm_data[taxon_id].append(taxon_data)
            imported_rows.append(row)

        elif "taxonomy" in types and not_blank(
            "alt_taxon_id", processed_data["taxonomy"], blanks
        ):
            without_ids[processed_data["taxonomy"]["alt_taxon_id"]].append(
                processed_data
            )
            taxon_asm_data[processed_data["taxonomy"]["alt_taxon_id"]].append(
                taxon_data
            )
            failed_rows[processed_data["taxonomy"]["alt_taxon_id"]].append(row)
        else:
            row_rank = None
            for rank in ranks:
                if not_blank(rank, processed_data["taxonomy"], blanks):
                    row_rank = rank
                    without_ids[processed_data["taxonomy"][rank]].append(processed_data)
                    taxon_asm_data[processed_data["taxonomy"][rank]].append(taxon_data)
                    failed_rows[processed_data["taxonomy"][rank]].append(row)
                    break
            if row_rank is None:
                failed_rows["None"].append(row)


def index_taxon_records(es, taxonomy_name, opts, with_ids, blanks, types):
    """Index a taxon records."""
    taxon_template = taxon.index_template(taxonomy_name, opts)
    docs = add_names_and_attributes_to_taxa(
        es, dict(with_ids), opts, template=taxon_template, blanks=blanks
    )
    imported_taxa = defaultdict(list)
    index_stream(
        es,
        taxon_template["index_name"],
        summarise_imported_taxa(docs, imported_taxa),
        _op_type="update",
        dry_run=opts.get("dry-run", False),
        log=opts.get("log-es", True),
        chunk_size=opts.get("es-batch", 500),
    )
    write_imported_taxa(imported_taxa, opts, types=types)


def index_sample_records(
    es,
    taxonomy_name,
    opts,
    with_ids,
    blanks,
    taxon_types,
    taxon_asm_data,
    index_type="sample",
):
    """Index sample records."""
    sample_template = sample.index_template(taxonomy_name, opts, index_type=index_type)
    taxon_template = taxon.index_template(taxonomy_name, opts)
    docs = add_identifiers_and_attributes_to_entries(
        es,
        with_ids,
        opts,
        template=sample_template,
        taxon_template=taxon_template,
        blanks=blanks,
        index_type=index_type,
    )
    index_stream(
        es,
        sample_template["index_name"],
        docs,
        dry_run=opts.get("dry-run", False),
        log=opts.get("log-es", True),
        chunk_size=opts.get("es-batch", 500),
    )
    # index taxon-level attributes
    index_types(
        es,
        "taxon",
        {"attributes": taxon_types},
        opts,
        dry_run=opts.get("dry-run", False),
    )
    taxon_asm_with_ids = {
        taxon_id: taxon_asm_data[taxon_id] for taxon_id in with_ids.keys()
    }
    # print(taxon_asm_with_ids)
    taxon_docs = add_names_and_attributes_to_taxa(
        es,
        taxon_asm_with_ids,
        opts,
        template=taxon_template,
        blanks=blanks,
    )
    index_stream(
        es,
        taxon_template["index_name"],
        taxon_docs,
        _op_type="update",
        dry_run=opts.get("dry-run", False),
        log=opts.get("log-es", True),
        chunk_size=opts.get("es-batch", 500),
    )


def process_taxon_sample_records(
    es,
    taxonomy_name,
    opts,
    processed_rows,
    with_ids,
    blanks,
    taxon_asm_data,
    imported_rows,
    types,
    failed_rows,
    header,
    taxon_table,
    taxon_types,
):
    """Process taxon and sample records."""
    taxon_template = taxon.index_template(taxonomy_name, opts)
    without_ids = defaultdict(list)
    if "taxon-id-as-xref" in opts:
        id_map = translate_xrefs(
            es,
            index=taxon_template["index_name"],
            xrefs=list(processed_rows.keys()),
            source=opts["taxon-id-as-xref"],
        )
        updated_rows = defaultdict(list)
        for xref, taxon_id in id_map.items():
            updated_rows[taxon_id] = processed_rows[xref]
            del processed_rows[xref]
        unmatched = list(processed_rows.keys())
        for xref in unmatched:
            updated_rows["other"] += processed_rows[xref]
            del processed_rows[xref]
    else:
        updated_rows = processed_rows
    for taxon_id, rows in updated_rows.items():
        group_rows(
            taxon_id,
            rows,
            with_ids,
            without_ids,
            taxon_asm_data,
            imported_rows,
            types,
            failed_rows,
            blanks,
        )
    LOGGER.info("Found taxon IDs in %d entries", len(with_ids.keys()))
    spellings = {"spellcheck": {}, "synonym": {}}
    create_ids, without_ids = fix_missing_ids(
        es,
        opts,
        without_ids,
        types=types,
        taxon_template=taxon_template,
        failed_rows=failed_rows,
        imported_rows=imported_rows,
        with_ids=with_ids,
        blanks=blanks,
        header=header,
        spellings=spellings,
        taxon_table=taxon_table,
    )
    write_spellchecked_taxa(spellings, opts, types=types)
    if with_ids or create_ids:
        write_imported_rows(
            imported_rows, opts, types=types, header=header, label="imported"
        )
        LOGGER.info("Indexing %d entries", len(with_ids.keys()))
        if opts["index"] == "taxon":
            index_taxon_records(es, taxonomy_name, opts, with_ids, blanks, types)
        elif opts["index"] == "assembly":
            # TODO: keep track of taxon_id not found exceptions
            index_sample_records(
                es,
                taxonomy_name,
                opts,
                with_ids,
                blanks,
                taxon_types,
                taxon_asm_data,
                index_type="assembly",
            )
        elif opts["index"] == "sample":
            # TODO: keep track of taxon_id not found exceptions
            index_sample_records(
                es,
                taxonomy_name,
                opts,
                with_ids,
                blanks,
                taxon_types,
                taxon_asm_data,
                index_type="sample",
            )


def convert_features_to_docs(with_ids):
    """Convert features to docs."""
    for taxon_id, entries in with_ids.items():
        for entry in entries:
            entry["taxon_id"] = taxon_id
            del entry["taxon_names"]
            del entry["taxonomy"]
            del entry["taxon_attributes"]
            yield (f'feature-{entry["feature_id"]}', entry)


def index_feature_records(es, opts, taxonomy_name, with_ids, blanks):
    """Index a feature records."""
    feature_template = feature.index_template(taxonomy_name, opts)
    # TODO: allow for adding attributes to existing features
    # TODO: allow for elevating summary attributes to assembly/taxon
    docs = convert_features_to_docs(dict(with_ids))
    index_stream(
        es,
        feature_template["index_name"],
        docs,
        dry_run=opts.get("dry-run", False),
        log=opts.get("log-es", True),
        chunk_size=opts.get("es-batch", 500),
    )


def index_file(
    es,
    types,
    names,
    data,
    opts,
    *,
    taxon_table=None,
    shared_values=None,
    exclusions=None,
):
    """Index a file."""
    delimiters = {"csv": ",", "tsv": "\t"}
    # Increase the maximum field size limit for CSV reader
    maxInt = sys.maxsize
    while True:
        # guard against overflow errors
        try:
            csv.field_size_limit(maxInt)
            break
        except OverflowError:
            maxInt = int(maxInt / 10)
    rows = csv.reader(
        strip_comments(data, types),
        delimiter=delimiters[types["file"]["format"]],
        quotechar='"',
    )
    if "header" in types["file"] and types["file"]["header"]:
        header = next(rows)
        set_column_indices(types, header)
    else:
        header = None
    with_ids = defaultdict(list)
    taxon_asm_data = defaultdict(list)
    failed_rows = defaultdict(list)
    imported_rows = []
    blanks = set(opts.get("blank", ["NA", "N/A", "None"]))
    blanks.add("")
    blanks.add(None)
    print(f"Blank values: {blanks}")
    taxon_types = {}
    taxonomy_name = opts["taxonomy-source"].lower()
    LOGGER.info("Processing rows")
    processed_rows = defaultdict(list)
    for row in tqdm(rows, mininterval=int(opts.get("log-interval", 1))):
        try:
            processed_data, taxon_data, new_taxon_types = process_row(
                types,
                names,
                row,
                shared_values,
                blanks,
                index_type=opts["index"],
                exclusions=exclusions,
            )
        except Exception:
            print(format_exc())
            failed_rows["None"].append(row)
            continue
        if processed_data is None:
            continue
        taxon_types.update(new_taxon_types)
        if opts["index"] == "feature" and not_blank(
            "taxon_id", processed_data["taxonomy"], blanks
        ):
            with_ids[processed_data["taxonomy"]["taxon_id"]].append(processed_data)
        elif not_blank("_taxon_id", processed_data["taxonomy"], blanks):
            # if opts["taxon-id-as-xref"]:
            with_ids[processed_data["taxonomy"]["_taxon_id"]].append(processed_data)
            taxon_asm_data[processed_data["taxonomy"]["_taxon_id"]].append(taxon_data)
            imported_rows.append(row)
        else:
            tmp_taxon_id = "other"
            if not_blank("taxon_id", processed_data["taxonomy"], blanks):
                tmp_taxon_id = processed_data["taxonomy"]["taxon_id"]
            processed_rows[tmp_taxon_id].append((processed_data, taxon_data, row))
    if opts["index"] in ["taxon", "sample", "assembly"]:
        process_taxon_sample_records(
            es,
            taxonomy_name,
            opts,
            processed_rows,
            with_ids,
            blanks,
            taxon_asm_data,
            imported_rows,
            types,
            failed_rows,
            header,
            taxon_table,
            taxon_types,
        )
    elif opts["index"] == "feature":
        index_feature_records(es, opts, taxonomy_name, with_ids, blanks)


def index_taxon_sample(es, opts, index="taxon", *, dry_run=False, taxonomy_name):
    """Call taxon- or sample-specific indexing functions."""
    taxon_table = None
    if taxon_table is None and "taxon-lookup-in-memory" in opts:
        taxon_table = {
            "scientific": defaultdict(list),
            "any": defaultdict(list),
        }
        load_taxon_table(es, opts, taxonomy_name, taxon_table)
    data_dir = f"{index}-dir"
    if data_dir in opts:
        dir_path = opts[data_dir]
        for types_file in sorted(Path(dir_path).glob("*.names.yaml")):
            types, data, names, exclusions = validate_types_file(
                types_file, dir_path, es, index, opts
            )
            if "file" in types and "name" in types["file"]:
                LOGGER.info(f'Indexing {types["file"]["name"]}')
                index_types(es, index, types, opts, dry_run=dry_run)
                index_file(
                    es,
                    types,
                    names,
                    data,
                    {
                        **opts,
                        "index": index,
                        "index_types": index_types,
                    },
                    taxon_table=taxon_table,
                )
                if "tests" in types["file"]:
                    result = test_json_dir(
                        f'{dir_path}/{types["file"]["tests"]}',
                        opts["es-host"][0],
                        opts,
                    )
                    if result is False:
                        LOGGER.error("Failed tests")
                        exit(1)
                        # time.sleep(5)
        for types_file in sorted(Path(dir_path).glob("*.types.yaml")):
            types, data, names, exclusions = validate_types_file(
                types_file, dir_path, es, index, opts
            )
            LOGGER.info("Indexing types")
            index_types(es, index, types, opts, dry_run=dry_run)
            if "file" in types and "name" in types["file"]:
                LOGGER.info(f'Indexing {types["file"]["name"]}')
                index_file(
                    es,
                    types,
                    names,
                    data,
                    {
                        **opts,
                        "index": index,
                        "index_types": index_types,
                    },
                    taxon_table=taxon_table,
                    exclusions=exclusions,
                )
                if "tests" in types["file"]:
                    result = test_json_dir(
                        f'{dir_path}/{types["file"]["tests"]}',
                        opts["es-host"][0],
                        opts,
                    )
                    if result is False:
                        LOGGER.error("Failed tests")
                        exit(1)


def set_feature_types(types):
    """Set types for feature properties."""
    if "features" not in types:
        return
    defaults = {
        "feature_id": "keyword",
        "assembly_id": "keyword",
        "sample_id": "keyword",
        "taxon_id": "keyword",
        "primary_type": "keyword",
    }
    for key, value in defaults.items():
        if key in types["features"]:
            if not isinstance(types["features"][key], dict):
                types["features"][key] = {"default": types["features"][key]}
            types["features"][key]["type"] = value


def index_features(es, opts, *, dry_run=False):
    """Index assembly features."""
    index = "feature"
    data_dir = f"{index}-dir"
    if data_dir in opts:
        dir_path = opts[data_dir]
    stored_attributes = {}
    file_list = list_files(dir_path, "*.types.yaml")
    shared_values = defaultdict(dict)
    shared_values["_es"] = es
    shared_values["_opts"] = opts
    template = feature.index_template(opts["taxonomy-source"].lower(), opts)
    shared_values["_index"] = template["index_name"]
    shared_values["_index_type"] = index
    for types_file in file_list:
        types, data, names, exclusions = validate_types_file(
            types_file, dir_path, es, index, opts, attributes=stored_attributes
        )
        LOGGER.info("Indexing types")
        if "file" in types and "name" in types["file"]:
            LOGGER.info(f'Indexing {types["file"]["name"]}')
            if "features" in types:
                set_feature_types(types)
            index_types(es, index, types, opts, dry_run=dry_run)
            sleep(1)
            shared_values["_types"] = types
            index_file(
                es,
                types,
                names,
                data,
                {**opts, "index": index, "index_types": index_types},
                shared_values=shared_values,
            )
        elif "attributes" in types:
            stored_attributes = {**stored_attributes, **types["attributes"]}


def main(args):
    """Index files."""
    options = config("index", **args)

    # Start Elasticsearch
    es = es_functions.launch_es(options["index"])

    # Post search scripts
    with tolog.DisableLogger():
        hub.post_search_scripts(es)

    taxonomy_name = options["index"]["taxonomy-source"].lower()
    dry_run = options["index"].get("dry-run", False)
    for index in ["taxon", "sample", "assembly"]:
        index_taxon_sample(
            es,
            options["index"],
            index=index,
            dry_run=dry_run,
            taxonomy_name=taxonomy_name,
        )

    if "feature-dir" in options["index"]:
        index_features(es, options["index"], dry_run=dry_run)

    if "file" in options["index"]:
        index_files(es, options["index"]["file"], taxonomy_name, options["index"])
    elif "file-metadata" in options["index"]:
        index_metadata(
            es,
            options["index"]["file-metadata"],
            taxonomy_name,
            options["index"],
            dry_run=dry_run,
        )


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index("index") + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    args["taxon-preload"] = True
    main(args)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""NCBI functions."""

import contextlib
import csv
import gzip
import json
import math
import os
import re
import subprocess
from collections import Counter
from collections import defaultdict
from itertools import groupby
from itertools import accumulate
from operator import itemgetter
from urllib.error import ContentTooShortError

import xml.etree.ElementTree as ET

import ujson
from Bio import SeqIO
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog
from tqdm import tqdm

from .geo import degrees_to_decimal

LOGGER = tolog.logger(__name__)

REFSEQ_FTP = "https://ftp.ncbi.nlm.nih.gov/refseq/release"


def refseq_listing(collection):
    """Fetch a directory listing for a RefSeq collection."""
    pattern = re.compile(r"(\w+\.\d+\.genomic\.gbff\.gz)")
    url = f"{REFSEQ_FTP}/{collection}"
    for _ in range(5):
        with contextlib.suppress(ContentTooShortError):
            html = tofetch.fetch_url(url)
            break
    listing = []
    for line in html.split("\n"):
        if match := pattern.search(line):
            listing.append(f"{url}/{match.group()}")
    return listing


def parse_references(entry, fields=None):
    """Parse references."""
    if fields is None:
        fields = {}
    submitted_year = re.compile(r"Submitted\s\(\d{2}-\w{3}-(\d{4})\)")
    published_year = re.compile(r"\s\((\d{4})\)[^\(]*$")
    for reference in entry.annotations["references"]:
        if reference.journal == "Unpublished":
            continue
        elif reference.journal.startswith("Submitted"):
            if "source_author" in fields:
                continue
            fields["source_year"] = submitted_year.search(reference.journal)[1]
        elif "source_author" in fields:
            continue
        else:
            fields["source_year"] = published_year.search(reference.journal)[1]
            if reference.title:
                fields["source_title"] = reference.title
            if reference.pubmed_id:
                fields["pubmed_id"] = reference.pubmed_id
        if reference.authors:
            fields["source_author"] = reference.authors
        elif reference.consrtm:
            fields["source_author"] = reference.consrtm
    return fields


def parse_xrefs(entry, fields=None):
    """Parse xrefs."""
    if fields is None:
        fields = {}
    if entry.dbxrefs:
        bioprojects = []
        biosamples = []
        for dbxref in entry.dbxrefs:
            with contextlib.suppress(ValueError):
                key, value = dbxref.split(":")
                if key == "BioProject":
                    bioprojects.append(value)
                elif key == "BioSample":
                    biosamples.append(value)
        if bioprojects:
            fields["bioproject"] = "; ".join(bioprojects)
        if biosamples:
            fields["biosample"] = "; ".join(biosamples)
    return fields


def parse_features(entry, fields):
    """Parse features."""
    if fields is None:
        fields = {}
    qualifiers = entry.features[0].qualifiers
    if "db_xref" in qualifiers:
        for xref in qualifiers["db_xref"]:
            key, value = xref.split(":")
            if key == "taxon":
                fields["taxon_id"] = value
    if "lat_lon" in qualifiers:
        fields["sample_location"] = qualifiers["lat_lon"]
    return fields


def reformat_date(string):
    """Change date string format."""
    months = {
        "JAN": "01",
        "FEB": "02",
        "MAR": "03",
        "APR": "04",
        "MAY": "05",
        "JUN": "06",
        "JUL": "07",
        "AUG": "08",
        "SEP": "09",
        "OCT": "10",
        "NOV": "11",
        "DEC": "12",
    }
    parts = re.split(r"[\:\-]", string)
    return f"{parts[2]}-{months[parts[1]]}-{parts[0].zfill(2)}"


def parse_flatfile(flatfile, organelle, opts):
    """Parse a GenBank flatfile."""
    data = []
    comment_re = re.compile(
        r"(?:derived|identical)\s(?:from|to)\s([\w\d]+).*COMPLETENESS: full length",
        re.DOTALL,
    )
    with gzip.open(flatfile, "rt") as fh:
        gb = SeqIO.parse(fh, "gb")
        for entry in tqdm(gb, mininterval=int(opts.get("log-interval", 1))):
            if (
                "refseq-root" in opts
                and opts["refseq-root"] not in entry.annotations["taxonomy"]
            ):
                continue
            fields = {"organelle": organelle}
            if entry.annotations["comment"]:
                if match := comment_re.search(entry.annotations["comment"]):
                    fields["genbank_accession"] = match[1]
                else:
                    continue
            parse_features(entry, fields)
            parse_references(entry, fields)
            fields["taxon"] = entry.annotations["organism"]
            fields["lineage"] = "; ".join(entry.annotations["taxonomy"])
            fields["assembly_id"] = entry.id
            fields["refseq_accession"] = entry.id
            fields["last_updated"] = reformat_date(entry.annotations["date"])
            parse_xrefs(entry, fields)
            try:
                seqstr = str(entry.seq.upper())
            except Exception:
                LOGGER.warn("Unable to read sequence for %s", entry.id)
                continue
            counter = Counter(seqstr)
            length = len(seqstr)
            fields["n_percent"] = float("%.2f" % (counter["N"] / length * 100))
            if fields["n_percent"] == 100:
                continue
            gc_count = counter["G"] + counter["C"]
            fields["gc_percent"] = float(
                "%.2f" % (gc_count / (gc_count + counter["A"] + counter["T"]) * 100)
            )
            fields["assembly_span"] = length
            data.append(fields)
    return data


def parse_listing(listing, collection, opts):
    """Parse all URLs in a directory listing."""
    parsed = []
    for url in listing:
        LOGGER.info("Fetching %s", url)
        flatfile = tofetch.fetch_tmp_file(url)
        LOGGER.info("Parsing %s", url)
        parsed += parse_flatfile(flatfile, collection, opts)
    return parsed


def refseq_organelle_parser(collections, opts, *, types=None, names=None):
    """Fetch and parse RefSeq organelle collections."""
    parsed = []
    if isinstance(collections, tuple):
        for collection in collections:
            listing = refseq_listing(collection)
            parsed += parse_listing(listing, collection, opts)
    else:
        listing = refseq_listing(collections)
        parsed += parse_listing(listing, collections, opts)
    # parsed = parse_flatfile(
    #     "mitochondrion.2.genomic.gbff.gz",
    #     "mitochondrion",
    #     opts,
    # )
    return parsed


def fetch_sequences_report(accession):
    """Fetch a sequence report from NCBI datasets."""
    result = subprocess.run(
        [
            "datasets",
            "summary",
            "genome",
            "accession",
            accession,
            "--report",
            "sequence",
        ],
        stdout=subprocess.PIPE,
        timeout=30,
    )
    return result.stdout


def set_organelle_values(seq, organelle, obj, organelle_name):
    """Set metadate values for an organelle"""
    organelle["genbankAssmAccession"] = seq[0]["genbank_accession"]
    organelle["totalSequenceLength"] = seq[0]["length"]
    organelle["gcPercent"] = seq[0]["gc_percent"]
    obj[f"{organelle_name}AssemblySpan"] = organelle["totalSequenceLength"]
    obj[f"{organelle_name}GcPercent"] = organelle["gcPercent"]
    obj[f"{organelle_name}Accession"] = seq[0]["genbank_accession"]


def add_organelle_entries(obj, organelles):
    """Add entries for co-assembled organelles."""
    obj["organelles"] = []
    for seq in organelles.values():
        try:
            organelle = {
                k: obj[k]
                for k in obj.keys()
                & {
                    "taxId",
                    "organismName",
                    "commonName",
                    "releaseDate",
                    "submitter",
                    "bioProjectAccession",
                    "biosampleAccession",
                }
            }
            organelle["sourceAccession"] = obj["genbankAssmAccession"]
            organelle["organelle"] = seq[0]["assigned_molecule_location_type"]
            organelle_name = (
                "mitochondrion"
                if seq[0]["assigned_molecule_location_type"] == "Mitochondrion"
                else "plastid"
            )
            if len(seq) == 1 and seq[0]["role"] == "assembled-molecule":
                set_organelle_values(seq, organelle, obj, organelle_name)
            else:
                obj[f"{organelle_name}Scaffolds"] = ";".join(
                    [entry["genbank_accession"] for entry in seq]
                )
            obj["organelles"].append(organelle)
        except Exception as err:
            LOGGER.warn(err)


def add_chromosome_entries(obj, chromosomes):
    """Add feature entries for assembled chromosomes."""
    obj["chromosomes"] = []
    for seq in chromosomes:
        obj["chromosomes"].append(
            {
                "sequence_id": seq["genbank_accession"],
                "start": 1,
                "end": seq["length"],
                "strand": 1,
                "length": seq["length"],
                "midpoint": math.round(seq["length"] / 2),
                "midpoint_proportion": 0.5,
                "seq_proportion": seq["length"] / obj["totalSequenceLength"],
            }
        )


def metricDates(obj):
    """Add date fields for assemblies with sufficient metrics."""
    accession = obj["genbankAssmAccession"]
    LOGGER.info(accession)
    span = int(obj["totalSequenceLength"])
    organelles = defaultdict(list)
    try:
        report = fetch_sequences_report(accession)
        data = json.loads(report)
    except Exception as err:
        LOGGER.warn(err)
        return
    if "reports" not in data:
        return
    chromosomes = []
    assigned_span = 0
    for seq in data["reports"]:
        if seq["assembly_unit"] == "non-nuclear":
            organelles[seq["chr_name"]].append(seq)
        elif seq["assembly_unit"] == "Primary Assembly":
            if seq["assigned_molecule_location_type"] in [
                "Chromosome",
                "Linkage Group",
            ] and seq["role"] in ["assembled-molecule", "unlocalized-scaffold"]:
                assigned_span += seq["length"]
                if seq["role"] == "assembled-molecule":
                    chromosomes.append(seq)

    if organelles:
        add_organelle_entries(obj, organelles)
    contig_n50 = obj.get("contigN50", 0)
    scaffold_n50 = obj.get("scaffoldN50", 0)
    if not chromosomes:
        obj["assignedProportion"] = None
        return
    obj["assignedProportion"] = assigned_span / span
    # add_chromosome_entries(obj, chromosomes)
    standardCriteria = []
    if contig_n50 >= 1000000 and scaffold_n50 >= 10000000:
        standardCriteria.append("6.7")
    if obj["assignedProportion"] >= 0.9:
        if contig_n50 >= 1000000:
            standardCriteria.append("6.C")
        elif scaffold_n50 < 1000000 and contig_n50 >= 100000:
            standardCriteria.append("5.C")
        elif scaffold_n50 < 10000000 and contig_n50 >= 100000:
            standardCriteria.append("5.6")
    if standardCriteria:
        obj["ebpStandardDate"] = obj["releaseDate"]
        obj["ebpStandardCriteria"] = ";".join(standardCriteria)


def get_biosample_attributes(biosample):
    """Extract attributes from a biosample."""
    obj = {}
    attributes = biosample.get("attributes", [])
    for entry in attributes:
        try:
            if entry["value"].lower().startswith("not "):
                continue
            obj[entry["name"]] = entry["value"]
        except KeyError:
            continue
    return obj


def parse_ncbi_datasets_record(record, parsed, previous):
    """Parse a single NCBI datasets record."""
    organism = record.get("organism", {})
    obj = {
        key: organism.get(key, "None")
        for key in ("taxId", "organismName", "commonName")
    }
    obj["organelle"] = "Nucleus"
    obj["genbankAssmAccession"] = record["accession"]
    if "pairedAccession" in record:
        if record["pairedAccession"].startswith("GCF_"):
            obj["refseqAssmAccession"] = record["pairedAccession"]
        else:
            obj["refseqAssmAccession"] = record["accession"]
            obj["genbankAssmAccession"] = record["pairedAccession"]

    if obj["genbankAssmAccession"] in parsed:
        obj = parsed[obj["genbankAssmAccession"]]
    assemblyInfo = record.get("assemblyInfo", {})
    annotationInfo = record.get("annotationInfo", {})
    if assemblyInfo.get("atypical", False):
        return
    for key in (
        "assemblyLevel",  #
        "assemblyName",  #
        "assemblyType",  #
        # "biosampleAccession",  # biosample.accession
        # "pairedAccession",  # refseqAssmAccession
        "refseqCategory",  #
        "releaseDate",  # submissionDate
        "submitter",  #
    ):
        obj[key] = assemblyInfo.get(key, annotationInfo.get(key, "None"))
    if (
        obj["genbankAssmAccession"] in previous
        and obj["releaseDate"] == previous[obj["genbankAssmAccession"]]["releaseDate"]
    ):
        parsed[obj["genbankAssmAccession"]] = previous[obj["genbankAssmAccession"]]
        if "organelles" in previous[obj["genbankAssmAccession"]]:
            for organelle in previous[obj["genbankAssmAccession"]]["organelles"]:
                parsed[organelle["genbankAssmAccession"]] = organelle
        return
    if "refseqAssmAccession" not in obj or obj["refseqAssmAccession"] == "na":
        obj["refseqAssmAccession"] = "None"
    if obj["refseqCategory"] != "None":
        obj["primaryValue"] = 1
    if annotationInfo:
        annot = {
            f"annotation{key.capitalize()}": annotationInfo.get(key, "None")
            for key in ("name", "releaseDate", "reportUrl", "source")  # sourceDatabase
        }

        if annot and "stats" in annotationInfo:
            geneCounts = annotationInfo["stats"].get("geneCounts", "None")
            for key in ("nonCoding", "proteinCoding", "pseudogene", "total"):
                annot[f"geneCount{key.capitalize()}"] = geneCounts.get(key, "None")
            obj.update(annot)
    bioprojects = []
    for lineage in assemblyInfo.get("bioprojectLineage", []):
        if "bioprojects" in lineage:
            bioprojects.extend(
                bioproject["accession"] for bioproject in lineage["bioprojects"]
            )

    if not bioprojects and "bioprojects" in assemblyInfo.get("biosample", {}):
        bioprojects.extend(
            bioproject["accession"]
            for bioproject in assemblyInfo["biosample"]["bioprojects"]
        )

    obj["bioProjectAccession"] = ";".join(bioprojects) if bioprojects else "None"
    biosample = assemblyInfo.get("biosample", {})
    obj["biosampleAccession"] = biosample.get("accession", "None")
    biosampleAttributes = get_biosample_attributes(biosample)
    for key in ("isolate", "sex"):
        obj[key] = biosampleAttributes.get(key, "None")
    assemblyStats = record.get("assemblyStats", {})
    obj.update(assemblyStats)
    if (
        obj["assemblyLevel"] in ["Chromosome", "Complete Genome"]
        and "assignedProportion" not in obj
    ):
        metricDates(obj)
    wgsInfo = record.get("wgsInfo", {})
    for key in ("masterWgsUrl", "wgsContigsUrl", "wgsProjectAccession"):
        obj[key] = wgsInfo.get(key, "None")
    if obj.get("genbankAssmAccession", "None") in parsed:
        for key, value in obj.items():
            if value != "None":
                parsed[obj["genbankAssmAccession"]][key] = value
    elif obj["genbankAssmAccession"].startswith("GCA_"):
        parsed[obj["genbankAssmAccession"]] = obj
        if "organelles" in obj:
            for organelle in obj["organelles"]:
                if "genbankAssmAccession" in organelle:
                    parsed[organelle["genbankAssmAccession"]] = organelle


def parse_ncbi_datasets_sample(record, parsed, previous):
    """Parse sample information from a single NCBI datasets record."""
    organism = record.get("organism", {})
    obj = {
        key: organism.get(key, "None")
        for key in ("taxId", "organismName", "commonName", "infraspecificNames")
    }
    if obj["infraspecificNames"] != "None":
        for key in ("sex", "isolate"):
            obj[key] = obj["infraspecificNames"].get(key, "None")
    del obj["infraspecificNames"]
    obj["genbankAssmAccession"] = record["accession"]
    if "pairedAccession" in record:
        if record["pairedAccession"].startswith("GCF_"):
            obj["refseqAssmAccession"] = record["pairedAccession"]
        else:
            obj["refseqAssmAccession"] = record["accession"]
            obj["genbankAssmAccession"] = record["pairedAccession"]
    assemblyInfo = record.get("assemblyInfo", {})
    annotationInfo = record.get("annotationInfo", {})
    if assemblyInfo.get("atypical", False):
        return

    for key in (
        "assemblyLevel",
        "refseqCategory",
        "submitter",
    ):
        obj[key] = assemblyInfo.get(key, annotationInfo.get(key, "None"))
    if "refseqAssmAccession" not in obj or obj["refseqAssmAccession"] == "na":
        obj["refseqAssmAccession"] = "None"
    elif obj["refseqCategory"] != "None":
        obj["primaryValue"] = 1
    bioprojects = []
    for lineage in assemblyInfo.get("bioprojectLineage", []):
        if "bioprojects" in lineage:
            bioprojects.extend(
                bioproject["accession"] for bioproject in lineage["bioprojects"]
            )

    if not bioprojects and "bioprojects" in assemblyInfo.get("biosample", {}):
        bioprojects.extend(
            bioproject["accession"]
            for bioproject in assemblyInfo["biosample"]["bioprojects"]
        )

    obj["bioProjectAccession"] = ";".join(bioprojects) if bioprojects else "None"
    biosample = assemblyInfo.get("biosample", {})
    obj["biosampleAccession"] = biosample.get("accession", "None")
    biosampleAttributes = get_biosample_attributes(biosample)

    biosampleAttributes = {
        entry["name"]: entry.get("value", "None")
        for entry in biosample.get("attributes", [])
    }
    for key in ("estimated_size", "geo_loc_name", "num_replicons", "ploidy"):
        obj[key] = biosampleAttributes.get(key)
    obj["latitude"] = degrees_to_decimal(
        biosampleAttributes.get("geographic location (latitude)")
    )
    obj["longitude"] = degrees_to_decimal(
        biosampleAttributes.get("geographic location (longitude)")
    )
    obj["elevation"] = biosampleAttributes.get("geographic location (elevation)")
    parsed[obj["biosampleAccession"]] = obj


def load_genome_tsv(file):
    """Load rows from a TSV file."""
    rows = {}
    with open(file, newline="") as tsv_file:
        reader = csv.DictReader(tsv_file, delimiter="\t")
        for row in reader:
            if "sourceAccession" in row:
                if row["sourceAccession"] == "None":
                    row["organelles"] = []
                    rows[row["genbankAssmAccession"]] = row
                else:
                    rows[row["sourceAccession"]]["organelles"].append(row)
    return rows


def ncbi_genome_parser(params, opts, *, types=None, names=None):
    """Parse NCBI Datasets genome report."""
    parsed = {}
    jsonl = opts[f"ncbi-datasets-{params}"]
    if not jsonl.endswith(".jsonl"):
        jsonl = f"{jsonl}/ncbi_dataset/data/assembly_data_report.jsonl"
    outfile = opts["outfile"]
    previous = load_genome_tsv(outfile) if os.path.isfile(outfile) else {}
    with tofile.open_file_handle(jsonl) as report:
        for line in report:
            record = ujson.loads(line)
            if params == "sample":
                parse_ncbi_datasets_sample(record, parsed, previous)
            else:
                parse_ncbi_datasets_record(record, parsed, previous)
    return list(parsed.values())


def split_chunks(values, split_val):
    """Split a list of values into chunks at a certain value."""
    index = 0

    def chunk_index(val):
        nonlocal index
        if val == split_val:
            index += 1
        return index

    return groupby(values, chunk_index)


def read_exp_xml(node, obj):
    """Read values from ExpXml section."""
    for child in node:
        tag = child.tag
        if tag == "Bioproject":
            obj["bioproject"] = child.text
        elif tag == "Biosample":
            obj["biosample"] = child.text
        elif tag == "Organism":
            obj["taxon_id"] = child.get("taxid")
        elif tag == "Experiment":
            obj["sra_accession"] = child.get("acc")
        elif tag == "Summary":
            obj["platform"] = child.findtext("Platform")
        elif tag == "Library_descriptor":
            obj["library_source"] = child.findtext("LIBRARY_SOURCE").lower()


def read_runs(node, obj):
    """Read values from Runs section."""
    if "runs" not in obj:
        obj["runs"] = []
    for child in node:
        obj["runs"].append(
            {"accession": child.get("acc"), "reads": child.get("total_spots")}
        )


def parse_sra_xml(xml_file):
    """Parse an SRA xml file."""
    rows = []
    with open(xml_file, "r", encoding="utf8") as container_file:
        for doc_num, doc in split_chunks(
            container_file, '<?xml version="1.0" encoding="UTF-8" ?>\n'
        ):
            LOGGER.info(f"processing sub-document #{doc_num}")
            LOGGER.info(len(rows))
            lines = list(doc)
            try:
                root = ET.fromstringlist(lines)
            except Exception:
                continue
            for doc_summary in root.iter("DocumentSummary"):
                obj = {}
                for child in doc_summary:
                    tag = child.tag
                    if tag == "CreateDate":
                        obj["date"] = child.text
                    elif tag == "ExpXml":
                        read_exp_xml(child, obj)
                    elif tag == "Runs":
                        read_runs(child, obj)
                    continue
                rows.append(obj)
    return rows


def group_by_taxon(rows, grouped=None):
    """
    Group SRA runs by taxon.

    Keep the most recent 10 rows only.
    """
    if grouped is None:
        grouped = defaultdict(lambda: {"count": 0, "reads": 0, "runs": []})
    for obj in sorted(rows, key=itemgetter("date")):
        try:
            taxon_id = obj["taxon_id"]
        except Exception:
            continue
        for run in obj["runs"]:
            try:
                row = {
                    **obj,
                    "run_accession": run["accession"],
                    "reads": int(run["reads"]),
                }
            except Exception:
                continue
            row.pop("runs")
            grouped[taxon_id]["runs"].insert(0, row)
            grouped[taxon_id]["count"] += 1
            grouped[taxon_id]["reads"] += row["reads"]
            if len(grouped[taxon_id]["runs"]) > 10:
                grouped[taxon_id]["runs"].pop()
    rows = [
        {
            "taxon_id": taxon_id,
            "sra_accession": ";".join([item["sra_accession"] for item in obj["runs"]]),
            "run_accession": ";".join([item["run_accession"] for item in obj["runs"]]),
            "library_source": ";".join(
                [item["library_source"] for item in obj["runs"]]
            ),
            "platform": ";".join([item["platform"] for item in obj["runs"]]),
            "reads": ";".join([str(item["reads"]) for item in obj["runs"]]),
            "total_reads": obj["reads"],
            "total_runs": obj["count"],
        }
        for taxon_id, obj in grouped.items()
    ]
    return rows


def load_sra_tsv(file):
    """Load rows from a TSV file."""
    grouped = defaultdict(lambda: {"count": 0, "reads": 0, "runs": []})
    with open(file, newline="") as tsv_file:
        reader = csv.DictReader(tsv_file, delimiter="\t")
        for row in reader:
            taxon_id = row["taxon_id"]
            total_runs = row["total_runs"]
            total_reads = row["total_reads"]
            grouped[taxon_id]["count"] = int(total_runs)
            grouped[taxon_id]["reads"] = int(total_reads)
            for key in (
                "run_accession",
                "sra_accession",
                "library_source",
                "platform",
                "reads",
            ):
                row[key] = row[key].split(";")
            for index, run_accession in enumerate(row["run_accession"]):
                grouped[taxon_id]["runs"].append(
                    {
                        "run_accession": run_accession,
                        "sra_accession": row["sra_accession"][index],
                        "library_source": row["library_source"][index],
                        "platform": row["platform"][index],
                        "reads": int(row["reads"][index]),
                    }
                )
    return grouped


def sra_parser(params, opts, *, types=None, names=None):
    """Parse SRA efetch xml."""
    outfile = opts["outfile"]
    # Load older runs from tsv
    previous = load_sra_tsv(outfile) if os.path.isfile(outfile) else None
    xml_file = opts["sra"]
    # Load newer runs from xml
    rows = parse_sra_xml(xml_file)
    return group_by_taxon(rows, grouped=previous)
